@@ -54,14 +54,17 @@ struct near_adapter {
 	near_bool_t constant_poll;
 	near_bool_t dep_up;
 
-	GHashTable *targets;
-	struct near_target *link;
-	int sock;
+	GHashTable *tags;
+	struct near_tag *tag_link;
+	int tag_sock;
+
+	GHashTable *devices;
+	struct near_device *device_link;
+	int device_sock;
 
 	GIOChannel *channel;
 	guint watch;
 	GList *ioreq_list;
-	GList *ndef_q;
 
 	guint presence_timeout;
 };
@@ -91,11 +94,18 @@ static void free_adapter(gpointer data)
 	g_free(adapter);
 }
 
-static void free_target(gpointer data)
+static void free_tag(gpointer data)
 {
-	struct near_target *target = data;
+	struct near_tag *tag = data;
 
-	__near_target_remove(target);
+	__near_tag_remove(tag);
+}
+
+static void free_device(gpointer data)
+{
+	struct near_device *device = data;
+
+	__near_device_remove(device);
 }
 
 static void polling_changed(struct near_adapter *adapter)
@@ -110,11 +120,18 @@ static int adapter_start_poll(struct near_adapter *adapter)
 {
 	int err;
 
-	if (g_hash_table_size(adapter->targets) > 0) {
-		DBG("Clearing targets");
+	if (g_hash_table_size(adapter->tags) > 0) {
+		DBG("Clearing tags");
 
-		g_hash_table_remove_all(adapter->targets);
-		__near_adapter_target_changed(adapter->idx);
+		g_hash_table_remove_all(adapter->tags);
+		__near_adapter_tags_changed(adapter->idx);
+	}
+
+	if (g_hash_table_size(adapter->devices) > 0) {
+		DBG("Clearing devices");
+
+		g_hash_table_remove_all(adapter->devices);
+		__near_adapter_devices_changed(adapter->idx);
 	}
 
 	err = __near_netlink_start_poll(adapter->idx, adapter->protocols);
@@ -191,45 +208,90 @@ static void append_protocols(DBusMessageIter *iter, void *user_data)
 	}
 }
 
-static void append_target_path(gpointer key, gpointer value, gpointer user_data)
+static void append_tag_path(gpointer key, gpointer value, gpointer user_data)
 {
-	struct near_target *target = value;
+	struct near_tag *tag = value;
 	DBusMessageIter *iter = user_data;
-	const char *target_path;
+	const char *tag_path;
 
-	target_path = __near_target_get_path(target);
-	if (target_path == NULL)
+	tag_path = __near_tag_get_path(tag);
+	if (tag_path == NULL)
 		return;
 
-	DBG("%s", target_path);
+	DBG("%s", tag_path);
 
 	dbus_message_iter_append_basic(iter, DBUS_TYPE_OBJECT_PATH,
-							&target_path);
+							&tag_path);
 }
 
-static void append_targets(DBusMessageIter *iter, void *user_data)
+static void append_tags(DBusMessageIter *iter, void *user_data)
 {
 	struct near_adapter *adapter = user_data;
 
 	DBG("");
 
-	g_hash_table_foreach(adapter->targets, append_target_path, iter);
+	g_hash_table_foreach(adapter->tags, append_tag_path, iter);
 }
 
-void __near_adapter_target_changed(uint32_t adapter_idx)
+static void append_device_path(gpointer key, gpointer value, gpointer user_data)
+{
+	struct near_device *device = value;
+	DBusMessageIter *iter = user_data;
+	const char *device_path;
+
+	device_path = __near_device_get_path(device);
+	if (device_path == NULL)
+		return;
+
+	DBG("%s", device_path);
+
+	dbus_message_iter_append_basic(iter, DBUS_TYPE_OBJECT_PATH,
+							&device_path);
+}
+
+static void append_devices(DBusMessageIter *iter, void *user_data)
+{
+	struct near_adapter *adapter = user_data;
+
+	DBG("");
+
+	g_hash_table_foreach(adapter->devices, append_device_path, iter);
+}
+
+void __near_adapter_tags_changed(uint32_t adapter_idx)
 {
 	struct near_adapter *adapter;
 
+	DBG("");
+
 	adapter = g_hash_table_lookup(adapter_hash,
 				GINT_TO_POINTER(adapter_idx));
 	if (adapter == NULL)
 		return;
 
 	near_dbus_property_changed_array(adapter->path,
-				NFC_ADAPTER_INTERFACE, "Targets",
-				DBUS_TYPE_OBJECT_PATH, append_targets,
+				NFC_ADAPTER_INTERFACE, "Tags",
+				DBUS_TYPE_OBJECT_PATH, append_tags,
 				adapter);
 }
+
+void __near_adapter_devices_changed(uint32_t adapter_idx)
+{
+	struct near_adapter *adapter;
+
+	DBG("");
+
+	adapter = g_hash_table_lookup(adapter_hash,
+				GINT_TO_POINTER(adapter_idx));
+	if (adapter == NULL)
+		return;
+
+	near_dbus_property_changed_array(adapter->path,
+				NFC_ADAPTER_INTERFACE, "Devices",
+				DBUS_TYPE_OBJECT_PATH, append_devices,
+				adapter);
+}
+
 static DBusMessage *get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -256,8 +318,11 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	near_dbus_dict_append_array(&dict, "Protocols",
 				DBUS_TYPE_STRING, append_protocols, adapter);
 
-	near_dbus_dict_append_array(&dict, "Targets",
-				DBUS_TYPE_OBJECT_PATH, append_targets, adapter);
+	near_dbus_dict_append_array(&dict, "Tags",
+				DBUS_TYPE_OBJECT_PATH, append_tags, adapter);
+
+	near_dbus_dict_append_array(&dict, "Devices",
+				DBUS_TYPE_OBJECT_PATH, append_devices, adapter);
 
 	near_dbus_dict_close(&array, &dict);
 
@@ -336,88 +401,13 @@ static DBusMessage *stop_poll(DBusConnection *conn,
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
-static int __push_ndef_queue(struct near_adapter *adapter,
-					struct near_ndef_message *ndef)
-{
-	if (adapter == NULL || ndef == NULL)
-		return -EINVAL;
-
-	adapter->ndef_q = g_list_append(adapter->ndef_q, ndef);
-
-	return 0;
-}
-
-static struct near_ndef_message *
-__pop_ndef_queue(struct near_adapter *adapter)
-{
-	GList *list;
-	struct near_ndef_message *ndef;
-
-	if (adapter == NULL)
-		return NULL;
-
-	list = g_list_first(adapter->ndef_q);
-	if (list == NULL)
-		return NULL;
-
-	ndef = list->data;
-	if (ndef != NULL)
-		adapter->ndef_q = g_list_remove(adapter->ndef_q, ndef);
-
-	return ndef;
-}
-
-static int __publish_text_record(DBusMessage *msg, void *data)
-{
-	DBusMessageIter iter, arr_iter;
-	struct near_ndef_message *ndef;
-	char *cod = NULL, *lang = NULL, *rep = NULL;
-
-	DBG("");
-
-	dbus_message_iter_init(msg, &iter);
-	dbus_message_iter_recurse(&iter, &arr_iter);
-
-	while (dbus_message_iter_get_arg_type(&arr_iter) !=
-					DBUS_TYPE_INVALID) {
-		const char *key;
-		DBusMessageIter ent_iter;
-		DBusMessageIter var_iter;
-
-		dbus_message_iter_recurse(&arr_iter, &ent_iter);
-		dbus_message_iter_get_basic(&ent_iter, &key);
-		dbus_message_iter_next(&ent_iter);
-		dbus_message_iter_recurse(&ent_iter, &var_iter);
-
-		switch (dbus_message_iter_get_arg_type(&var_iter)) {
-		case DBUS_TYPE_STRING:
-			if (g_strcmp0(key, "Encoding") == 0)
-				dbus_message_iter_get_basic(&var_iter, &cod);
-			else if (g_strcmp0(key, "Language") == 0)
-				dbus_message_iter_get_basic(&var_iter, &lang);
-			else if (g_strcmp0(key, "Representation") == 0)
-				dbus_message_iter_get_basic(&var_iter, &rep);
-
-			break;
-		}
-
-		dbus_message_iter_next(&arr_iter);
-	}
-
-	ndef = near_ndef_prepare_text_record(cod, lang, rep);
-	if (ndef == NULL)
-		return -EINVAL;
-
-	return __push_ndef_queue(data, ndef);
-}
-
 static void tag_present_cb(uint32_t adapter_idx, uint32_t target_idx,
 								int status);
 
 static gboolean check_presence(gpointer user_data)
 {
 	struct near_adapter *adapter = user_data;
-	struct near_target *target;
+	struct near_tag *tag;
 	int err;
 
 	DBG("");
@@ -425,20 +415,24 @@ static gboolean check_presence(gpointer user_data)
 	if (adapter == NULL)
 		return FALSE;
 
-	target = adapter->link;
-	if (target == NULL)
-		return FALSE;
+	tag = adapter->tag_link;
+	if (tag == NULL)
+		goto out_err;
 
-	err = __near_tag_check_presence(target, tag_present_cb);
+	err = __near_tag_check_presence(tag, tag_present_cb);
 	if (err < 0) {
 		DBG("Could not check target presence");
-
-		near_adapter_disconnect(adapter->idx);
-		if (adapter->constant_poll == TRUE)
-			adapter_start_poll(adapter);
+		goto out_err;
 	}
 
 	return FALSE;
+
+out_err:
+	near_adapter_disconnect(adapter->idx);
+	if (adapter->constant_poll == TRUE)
+		adapter_start_poll(adapter);
+
+	return FALSE;
 }
 
 static void tag_present_cb(uint32_t adapter_idx, uint32_t target_idx,
@@ -454,7 +448,7 @@ static void tag_present_cb(uint32_t adapter_idx, uint32_t target_idx,
 		return;
 
 	if (status < 0) {
-		DBG("Target is gone");
+		DBG("Tag is gone");
 
 		near_adapter_disconnect(adapter->idx);
 		if (adapter->constant_poll == TRUE)
@@ -468,216 +462,18 @@ static void tag_present_cb(uint32_t adapter_idx, uint32_t target_idx,
 					check_presence, adapter);
 }
 
-static char *get_uri_field(DBusMessage *msg)
-{
-	DBusMessageIter iter, arr_iter;
-	char *uri = NULL;
-
-	DBG("");
-
-	dbus_message_iter_init(msg, &iter);
-	dbus_message_iter_recurse(&iter, &arr_iter);
-
-	while (dbus_message_iter_get_arg_type(&arr_iter) !=
-					DBUS_TYPE_INVALID) {
-		const char *key;
-		DBusMessageIter ent_iter;
-		DBusMessageIter var_iter;
-
-		dbus_message_iter_recurse(&arr_iter, &ent_iter);
-		dbus_message_iter_get_basic(&ent_iter, &key);
-		dbus_message_iter_next(&ent_iter);
-		dbus_message_iter_recurse(&ent_iter, &var_iter);
-
-		switch (dbus_message_iter_get_arg_type(&var_iter)) {
-		case DBUS_TYPE_STRING:
-			if (g_strcmp0(key, "URI") == 0)
-				dbus_message_iter_get_basic(&var_iter, &uri);
-
-			break;
-		}
-
-		dbus_message_iter_next(&arr_iter);
-	}
-
-	return uri;
-}
-
-static int publish_uri_record(DBusMessage *msg, void *data)
-{
-	struct near_ndef_message *ndef;
-	char *uri = NULL;
-	const char *uri_prefix = NULL;
-	uint8_t id_len, i;
-	uint32_t uri_len;
-
-	DBG("");
-
-	uri = get_uri_field(msg);
-	if (uri == NULL)
-		return -EINVAL;
-
-	for (i = 1; i <= NFC_MAX_URI_ID; i++) {
-		uri_prefix = __near_ndef_get_uri_prefix(i);
-
-		if (uri_prefix != NULL &&
-				g_str_has_prefix(uri, uri_prefix) == TRUE)
-			break;
-	}
-
-	/* If uri_prefix is NULL then ID will be zero */
-	if (uri_prefix == NULL) {
-		i = 0;
-		id_len = 0;
-	} else
-		id_len = strlen(uri_prefix);
-
-	uri_len = strlen(uri) - id_len;
-	ndef = near_ndef_prepare_uri_record(i, uri_len,
-						(uint8_t *)(uri + id_len));
-	if (ndef == NULL)
-		return -EINVAL;
-
-	return __push_ndef_queue(data, ndef);
-}
-
-static int publish_sp_record(DBusMessage *msg, void *data)
-{
-	struct near_ndef_message *ndef;
-	char *uri = NULL;
-	const char *uri_prefix;
-	uint8_t id_len, i;
-	uint32_t uri_len;
-
-	DBG("");
-
-	/* Currently this funtion supports only mandatory URI record,
-	 * TODO: Other records support */
-	uri = get_uri_field(msg);
-	if (uri == NULL)
-		return -EINVAL;
-
-	for (i = 1; i <= NFC_MAX_URI_ID; i++) {
-		uri_prefix = __near_ndef_get_uri_prefix(i);
-
-		if (uri_prefix != NULL &&
-				g_str_has_prefix(uri, uri_prefix) == TRUE)
-			break;
-	}
-
-	if (uri_prefix == NULL) {
-		i = 0;
-		id_len = 0;
-	} else
-		id_len = strlen(uri_prefix);
-
-	uri_len = strlen(uri) - id_len;
-	ndef = near_ndef_prepare_smartposter_record(i, uri_len,
-						(uint8_t *)(uri + id_len));
-	if (ndef == NULL)
-		return -EINVAL;
-
-	return __push_ndef_queue(data, ndef);
-}
-
-static void __add_ndef_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
-{
-	struct near_adapter *adapter;
-
-	DBG("%d", status);
-
-	adapter = g_hash_table_lookup(adapter_hash,
-					GINT_TO_POINTER(adapter_idx));
-	if (adapter == NULL)
-		return;
-
-	adapter->presence_timeout =
-		g_timeout_add_seconds(CHECK_PRESENCE_PERIOD,
-					check_presence, adapter);
-}
-
-static DBusMessage *publish(DBusConnection *conn,
-				DBusMessage *msg, void *data)
-{
-	DBusMessageIter iter;
-	DBusMessageIter arr_iter;
-	struct near_adapter *adapter = data;
-
-	DBG("conn %p", conn);
-
-	dbus_message_iter_init(msg, &iter);
-	dbus_message_iter_recurse(&iter, &arr_iter);
-
-	while (dbus_message_iter_get_arg_type(&arr_iter) !=
-						DBUS_TYPE_INVALID) {
-		const char *key, *value;
-		DBusMessageIter ent_iter;
-		DBusMessageIter var_iter;
-
-		dbus_message_iter_recurse(&arr_iter, &ent_iter);
-		dbus_message_iter_get_basic(&ent_iter, &key);
-
-		if (g_strcmp0(key, "Type") != 0) {
-			dbus_message_iter_next(&arr_iter);
-			continue;
-		}
-
-		dbus_message_iter_next(&ent_iter);
-		dbus_message_iter_recurse(&ent_iter, &var_iter);
-
-		switch (dbus_message_iter_get_arg_type(&var_iter)) {
-		case DBUS_TYPE_STRING:
-			dbus_message_iter_get_basic(&var_iter, &value);
-
-			if (g_strcmp0(value, "Text") == 0) {
-				if (__publish_text_record(msg, adapter)	< 0)
-						goto error;
-
-				goto reply;
-			} else if (g_strcmp0(value, "URI") == 0) {
-				if (publish_uri_record(msg, adapter)	< 0)
-					goto error;
-
-				goto reply;
-
-			} else if (g_strcmp0(value, "SmartPoster") == 0) {
-				if (publish_sp_record(msg, adapter) < 0)
-					goto error;
-
-				goto reply;
-
-			} else {
-				DBG(" '%s' not supported", value);
-				goto error;
-			}
-
-				break;
-		}
-
-		dbus_message_iter_next(&arr_iter);
-	}
-
-error:
-	return g_dbus_create_error(msg, "org.neard.Error.InvalidArguments",
-							"Invalid arguments");
-
-reply:
-	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
-}
-
 static GDBusMethodTable adapter_methods[] = {
 	{ "GetProperties",     "",      "a{sv}", get_properties     },
 	{ "SetProperty",       "sv",    "",      set_property       },
 	{ "StartPoll",         "",      "",      start_poll         },
 	{ "StopPoll",          "",      "",      stop_poll          },
-	{ "Publish",         "a{sv}",   "",      publish            },
 	{ },
 };
 
 static GDBusSignalTable adapter_signals[] = {
 	{ "PropertyChanged",		"sv"	},
-	{ "TargetFound",		"o"	},
-	{ "TargetLost",			"o"	},
+	{ "TagFound",		        "o"	},
+	{ "TagLost",			"o"	},
 	{ }
 };
 
@@ -700,9 +496,13 @@ struct near_adapter * __near_adapter_create(uint32_t idx,
 	adapter->powered = powered;
 	adapter->constant_poll = near_setting_get_bool("ConstantPoll");
 	adapter->dep_up = FALSE;
-	adapter->targets = g_hash_table_new_full(g_direct_hash, g_direct_equal,
-							NULL, free_target);
-	adapter->sock = -1;
+	adapter->tags = g_hash_table_new_full(g_direct_hash, g_direct_equal,
+							NULL, free_tag);
+	adapter->tag_sock = -1;
+
+	adapter->devices = g_hash_table_new_full(g_direct_hash, g_direct_equal,
+							NULL, free_device);
+	adapter->device_sock = -1;
 
 	adapter->path = g_strdup_printf("%s/nfc%d", NFC_PATH, idx);
 
@@ -726,20 +526,7 @@ struct near_adapter *__near_adapter_get(uint32_t idx)
 	return g_hash_table_lookup(adapter_hash, GINT_TO_POINTER(idx));
 }
 
-int near_adapter_get_dep_state(uint32_t idx)
-{
-	struct near_adapter *adapter;
-
-	DBG("idx %d", idx);
-
-	adapter = g_hash_table_lookup(adapter_hash, GINT_TO_POINTER(idx));
-	if (adapter == NULL)
-		return -ENODEV;
-
-	return adapter->dep_up;
-}
-
-int near_adapter_set_dep_state(uint32_t idx, near_bool_t dep)
+int __near_adapter_set_dep_state(uint32_t idx, near_bool_t dep)
 {
 	struct near_adapter *adapter;
 
@@ -751,6 +538,9 @@ int near_adapter_set_dep_state(uint32_t idx, near_bool_t dep)
 
 	adapter->dep_up = dep;
 
+	if (dep == FALSE && adapter->constant_poll == TRUE)
+		adapter_start_poll(adapter);
+
 	return 0;
 }
 
@@ -785,31 +575,16 @@ void __near_adapter_remove(struct near_adapter *adapter)
 	g_hash_table_remove(adapter_hash, GINT_TO_POINTER(adapter->idx));
 }
 
-static int dep_link_up(uint32_t idx, uint32_t target_idx)
-{
-	return __near_netlink_dep_link_up(idx, target_idx,
-					NFC_COMM_ACTIVE, NFC_RF_INITIATOR);
-}
-
-static int dep_link_down(uint32_t idx)
-{
-	return __near_netlink_dep_link_down(idx);
-}
-
 static void tag_read_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
 {
 	struct near_adapter *adapter;
-	struct near_target *target;
-	struct near_ndef_message *ndef, *ndef_with_header = NULL;
-	uint16_t tag_type;
-	int err;
 
-	DBG("");
+	DBG("status %d", status);
 
 	adapter = g_hash_table_lookup(adapter_hash,
 					GINT_TO_POINTER(adapter_idx));
 	if (adapter == NULL)
-		goto out;
+		return;
 
 	if (status < 0) {
 		near_adapter_disconnect(adapter->idx);
@@ -819,107 +594,103 @@ static void tag_read_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
 		return;
 	}
 
-	__near_adapter_target_changed(adapter_idx);
+	__near_adapter_tags_changed(adapter_idx);
 
-	/* Check if adapter ndef queue has any ndef messages,
-	 * then write the ndef data on tag. */
-	if (g_list_length(adapter->ndef_q) == 0)
-		goto out;
+	adapter->presence_timeout =
+		g_timeout_add_seconds(CHECK_PRESENCE_PERIOD,
+					check_presence, adapter);
+}
 
-	target = g_hash_table_lookup(adapter->targets,
-					GINT_TO_POINTER(target_idx));
-	if (target == NULL)
-		goto out;
+static void device_read_cb(uint32_t adapter_idx, uint32_t target_idx,
+								int status)
+{
+	struct near_adapter *adapter;
 
-	ndef = __pop_ndef_queue(adapter);
-	if (ndef == NULL)
-		goto out;
+	DBG("status %d", status);
 
-	tag_type = __near_target_get_tag_type(target);
+	adapter = g_hash_table_lookup(adapter_hash,
+					GINT_TO_POINTER(adapter_idx));
+	if (adapter == NULL)
+		return;
 
-	/* Add NDEF header information depends upon tag type */
-	switch (tag_type) {
-	case NFC_PROTO_JEWEL:
-	case NFC_PROTO_MIFARE:
-		ndef_with_header = g_try_malloc0(sizeof(
-					struct near_ndef_message));
-		if (ndef_with_header == NULL)
-			goto out;
+	if (status < 0) {
+		if (adapter->device_link != NULL) {
+			__near_netlink_dep_link_down(adapter->idx);
+			adapter->device_link = NULL;
+		}
 
-		ndef_with_header->offset = 0;
-		ndef_with_header->length = ndef->length + 3;
-		ndef_with_header->data = g_try_malloc0(ndef->length + 3);
-		if (ndef_with_header->data == NULL)
-			goto out;
+		if (adapter->constant_poll == TRUE)
+			adapter_start_poll(adapter);
 
-		ndef_with_header->data[0] = TLV_NDEF;
-		ndef_with_header->data[1] = ndef->length;
-		memcpy(ndef_with_header->data + 2, ndef->data, ndef->length);
-		ndef_with_header->data[ndef->length + 2] = TLV_END;
-
-		break;
-
-	case NFC_PROTO_FELICA:
-		ndef_with_header = g_try_malloc0(sizeof(
-					struct near_ndef_message));
-		if (ndef_with_header == NULL)
-			goto out;
-
-		ndef_with_header->offset = 0;
-		ndef_with_header->length = ndef->length;
-		ndef_with_header->data = g_try_malloc0(
-						ndef_with_header->length);
-		if (ndef_with_header->data == NULL)
-			goto out;
-
-		memcpy(ndef_with_header->data, ndef->data, ndef->length);
-
-		break;
-
-	case NFC_PROTO_ISO14443:
-		ndef_with_header = g_try_malloc0(sizeof(
-					struct near_ndef_message));
-		if (ndef_with_header == NULL)
-			goto out;
-
-		ndef_with_header->offset = 0;
-		ndef_with_header->length = ndef->length + 2;
-		ndef_with_header->data = g_try_malloc0(ndef->length + 2);
-		if (ndef_with_header->data == NULL)
-			goto out;
-
-		ndef_with_header->data[0] = (uint8_t)(ndef->length >> 8);
-		ndef_with_header->data[1] = (uint8_t)(ndef->length);
-		memcpy(ndef_with_header->data + 2, ndef->data, ndef->length);
-
-		break;
-
-	default:
-		goto out;
+		return;
 	}
 
-	g_free(ndef->data);
-	g_free(ndef);
+	__near_adapter_devices_changed(adapter_idx);
+}
 
-	err = __near_tag_add_ndef(target, ndef_with_header, __add_ndef_cb);
+
+static int adapter_add_tag(struct near_adapter *adapter, uint32_t target_idx,
+			uint32_t protocols,
+			uint16_t sens_res, uint8_t sel_res,
+			uint8_t *nfcid, uint8_t nfcid_len)
+{
+	struct near_tag *tag;
+	uint32_t tag_type;
+	int err;
+
+	tag = __near_tag_add(adapter->idx, target_idx, protocols,
+				sens_res, sel_res,
+				nfcid, nfcid_len);
+	if (tag == NULL)
+		return -ENODEV;
+
+	g_hash_table_insert(adapter->tags, GINT_TO_POINTER(target_idx),
+								tag);
+
+	tag_type = __near_tag_get_type(tag);
+
+	err = near_adapter_connect(adapter->idx, target_idx, tag_type);
 	if (err < 0) {
-		g_free(ndef_with_header->data);
-		g_free(ndef_with_header);
-		goto out;
+		near_error("Could not connect");
+		return err;
 	}
 
-	return;
+	return __near_tag_read(tag, tag_read_cb);
+}
 
-out:
-	if (ndef_with_header != NULL) {
-		g_free(ndef_with_header->data);
-		g_free(ndef_with_header);
+static int adapter_add_device(struct near_adapter *adapter,
+				uint32_t target_idx,
+				uint8_t *nfcid, uint8_t nfcid_len)
+{
+	struct near_device *device;
+	int err;
+
+	device = __near_device_add(adapter->idx, target_idx, nfcid, nfcid_len);
+	if (device == NULL)
+		return -ENODEV;
+
+	g_hash_table_insert(adapter->devices, GINT_TO_POINTER(target_idx),
+								device);
+
+	/* For p2p, reading is listening for an incoming connection */
+	err = __near_device_listen(device, device_read_cb);
+	if (err < 0) {
+		near_error("Could not read device");
+		return err;
 	}
 
-	if (adapter != NULL)
-		adapter->presence_timeout =
-			g_timeout_add_seconds(CHECK_PRESENCE_PERIOD,
-						check_presence, adapter);
+	adapter->device_link = device;
+
+	if (adapter->dep_up == TRUE)
+		return 0;
+
+	err = __near_netlink_dep_link_up(adapter->idx, target_idx,
+					NFC_COMM_ACTIVE, NFC_RF_INITIATOR);
+
+	if (err < 0)
+		adapter->device_link = NULL;
+
+	return err;
 }
 
 int __near_adapter_add_target(uint32_t idx, uint32_t target_idx,
@@ -927,9 +698,6 @@ int __near_adapter_add_target(uint32_t idx, uint32_t target_idx,
 			uint8_t *nfcid, uint8_t nfcid_len)
 {
 	struct near_adapter *adapter;
-	struct near_target *target;
-	uint16_t tag_type;
-	int err;
 
 	DBG("idx %d", idx);
 
@@ -940,37 +708,13 @@ int __near_adapter_add_target(uint32_t idx, uint32_t target_idx,
 	adapter->polling = FALSE;
 	polling_changed(adapter);
 
-	/* TODO target reference */
-	target = __near_target_add(idx, target_idx, protocols,
+	if (protocols & NFC_PROTO_NFC_DEP_MASK)
+		return adapter_add_device(adapter, target_idx,
+						nfcid, nfcid_len);
+	else
+		return adapter_add_tag(adapter, target_idx, protocols,
 					sens_res, sel_res, nfcid, nfcid_len);
-	if (target == NULL)
-		return -ENODEV;
 
-	g_hash_table_insert(adapter->targets,
-			GINT_TO_POINTER(target_idx), target);	
-
-	tag_type = __near_target_get_tag_type(target);
-
-	if (tag_type != NFC_PROTO_NFC_DEP) {
-		err = near_adapter_connect(idx, target_idx, tag_type);
-		if (err < 0) {
-			near_error("Could not connect");
-			return err;
-		}
-
-		err = __near_tag_read(target, tag_read_cb);
-	} else {
-		/* For p2p, reading is service binding */
-		err = __near_tag_read(target, tag_read_cb);
-		if (err < 0) {
-			near_error("Could not read tag");
-			return err;
-		}
-
-		err = near_adapter_connect(idx, target_idx, tag_type);
-	}
-
-	return err;
 }
 
 int __near_adapter_remove_target(uint32_t idx, uint32_t target_idx)
@@ -983,9 +727,19 @@ int __near_adapter_remove_target(uint32_t idx, uint32_t target_idx)
 	if (adapter == NULL)
 		return -ENODEV;
 
-	g_hash_table_remove(adapter->targets, GINT_TO_POINTER(target_idx));
+	if (g_hash_table_remove(adapter->tags,
+			GINT_TO_POINTER(target_idx)) == TRUE) {      	
+		__near_adapter_tags_changed(idx);
 
-	__near_adapter_target_changed(idx);
+		return 0;
+	}
+
+	if (g_hash_table_remove(adapter->devices,
+			GINT_TO_POINTER(target_idx)) == TRUE) {      	
+		__near_adapter_devices_changed(idx);
+
+		return 0;
+	}
 
 	return 0;
 }
@@ -1035,6 +789,12 @@ static gboolean adapter_recv_event(GIOChannel *channel, GIOCondition condition,
 		near_error("Error while reading NFC bytes");
 
 		adapter_flush_rx(adapter, -EIO);
+
+		near_adapter_disconnect(adapter->idx);
+
+		adapter->presence_timeout =
+			g_timeout_add_seconds(2 * CHECK_PRESENCE_PERIOD,
+					      check_presence, adapter);
 		return FALSE;
 	}
 
@@ -1056,7 +816,7 @@ static gboolean adapter_recv_event(GIOChannel *channel, GIOCondition condition,
 int near_adapter_connect(uint32_t idx, uint32_t target_idx, uint8_t protocol)
 {
 	struct near_adapter *adapter;
-	struct near_target *target;
+	struct near_tag *tag;
 	struct sockaddr_nfc addr;
 	int err, sock;
 
@@ -1066,26 +826,14 @@ int near_adapter_connect(uint32_t idx, uint32_t target_idx, uint8_t protocol)
 	if (adapter == NULL)
 		return -ENODEV;
 
-	if (adapter->sock != -1)
+	if (adapter->tag_sock != -1)
 		return -EALREADY;
 
-	target = g_hash_table_lookup(adapter->targets,
+	tag = g_hash_table_lookup(adapter->tags,
 				GINT_TO_POINTER(target_idx));
-	if (target == NULL)
+	if (tag == NULL)
 		return -ENOLINK;
 
-	if (protocol == NFC_PROTO_NFC_DEP) {
-		err = dep_link_up(idx, target_idx);
-		if (err < 0)
-			return err;
-
-		adapter->link = target;
-
-		DBG("link %p", adapter->link);
-
-		return 0;
-	}
-
 	sock = socket(AF_NFC, SOCK_SEQPACKET, NFC_SOCKPROTO_RAW);
 	if (sock == -1)
 		return sock;
@@ -1101,11 +849,11 @@ int near_adapter_connect(uint32_t idx, uint32_t target_idx, uint8_t protocol)
 		return err;
 	}
 
-	adapter->sock = sock;
-	adapter->link = target;
+	adapter->tag_sock = sock;
+	adapter->tag_link = tag;
 
 	if (adapter->channel == NULL)
-		adapter->channel = g_io_channel_unix_new(adapter->sock);
+		adapter->channel = g_io_channel_unix_new(adapter->tag_sock);
 
 	g_io_channel_set_flags(adapter->channel, G_IO_FLAG_NONBLOCK, NULL);
 	g_io_channel_set_close_on_unref(adapter->channel, TRUE);
@@ -1130,24 +878,19 @@ int near_adapter_disconnect(uint32_t idx)
 	if (adapter == NULL)
 		return -ENODEV;
 
-	DBG("link %p", adapter->link);
+	DBG("link %p", adapter->tag_link);
 
-	if (adapter->link == NULL)
+	if (adapter->tag_link == NULL)
 		return -ENOLINK;
 
-	tag_type = __near_target_get_tag_type(adapter->link);
-	target_idx = __near_target_get_idx(adapter->link);
+	tag_type = __near_tag_get_type(adapter->tag_link);
+	target_idx = __near_tag_get_idx(adapter->tag_link);
 
 	DBG("tag type %d", tag_type);
 
 	__near_adapter_remove_target(adapter->idx, target_idx);
 
-	if (tag_type == NFC_PROTO_NFC_DEP) {
-		adapter->link = NULL;
-		return dep_link_down(idx);
-	}
-
-	if (adapter->sock == -1)
+	if (adapter->tag_sock == -1)
 		return -ENOLINK;
 
 	if (adapter->watch > 0) {
@@ -1156,9 +899,9 @@ int near_adapter_disconnect(uint32_t idx)
 	}
 
 	adapter->channel = NULL;
-	close(adapter->sock);
-	adapter->sock = -1;
-	adapter->link = NULL;
+	close(adapter->tag_sock);
+	adapter->tag_sock = -1;
+	adapter->tag_link = NULL;
 
 	return 0;
 }
@@ -1176,7 +919,7 @@ int near_adapter_send(uint32_t idx, uint8_t *buf, size_t length,
 	if (adapter == NULL)
 		return -ENODEV;
 
-	if (adapter->sock == -1 || adapter->link == NULL)
+	if (adapter->tag_sock == -1 || adapter->tag_link == NULL)
 		return -ENOLINK;
 
 	if (cb != NULL && adapter->watch != 0) {
@@ -1186,7 +929,7 @@ int near_adapter_send(uint32_t idx, uint8_t *buf, size_t length,
 
 		DBG("req %p cb %p data %p", req, cb, data);
 
-		req->target_idx = __near_target_get_idx(adapter->link);
+		req->target_idx = __near_tag_get_idx(adapter->tag_link);
 		req->cb = cb;
 		req->data = data;
 
@@ -1194,7 +937,7 @@ int near_adapter_send(uint32_t idx, uint8_t *buf, size_t length,
 			g_list_append(adapter->ioreq_list, req);
 	}
 
-	err = send(adapter->sock, buf, length, 0);
+	err = send(adapter->tag_sock, buf, length, 0);
 	if (err < 0)
 		goto out_err;
 
